@@ -1,14 +1,9 @@
 """
-OCL Flex Importer --
-Script that uses the OCL API to import multiple resource types from a JSON lines file.
-Configuration for individual resources can be set inline in the JSON. The bulk import
-API exposes an endpoint to submit a JSON lines for higher performance asynchronous processing
-directly on the OCL server. See oclapi Bulk Importing documentation for more information:
-https://github.com/OpenConceptLab/oclapi/wiki/Bulk-Importing
-
-NOTE: For batch imports of concepts or mappings into a single source, the server-side import
-script offers the highest performance alternative to this module. However, support for the
-server-side import will be discontinued as new functionality is added to the bulk import API.
+OCL Importer --
+Classes to import a list of resources into OCL. OclFlexImporter processes an import locally,
+submitting one resource at a time to the OCL API. OclBulkImporter submits an import script
+to an OCL server for higher performance asynchronous processing. See oclapi Bulk Importing
+documentation for more information: https://github.com/OpenConceptLab/oclapi/wiki/Bulk-Importing
 
 Verbosity settings:
 * 0 = show only responses from server
@@ -72,6 +67,14 @@ class InvalidObjectDefinition(OclImportError):
         self.message = message
 
 
+class UnsupportedActionType(OclImportError):
+    """ Exception raised for an unsupported action type """
+    def __init__(self, expression, message):
+        OclImportError.__init__(self)
+        self.expression = expression
+        self.message = message
+
+
 class OclImportResults(object):
     """ Class to capture and process the results of processing an import script """
 
@@ -180,7 +183,7 @@ class OclImportResults(object):
                 for status_code in self._results[root_key][action_type]:
                     try:
                         status_code_int = int(status_code)
-                    except:
+                    except Exception:
                         status_code_int = 0
                     if status_code == 'skip':
                         return True
@@ -451,14 +454,50 @@ class OclFlexImporter(object):
     DEFAULT_REFERENCE_BATCH_SIZE = 0
 
     # Constants for import action types
-    ACTION_TYPE_NEW = 'new'
+    ACTION_TYPE_DELETE = 'delete'
+    ACTION_TYPE_CREATE_OR_UPDATE = 'CREATE_OR_UPDATE'
+    ACTION_TYPE_CREATE = 'new'
     ACTION_TYPE_UPDATE = 'update'
     ACTION_TYPE_RETIRE = 'retire'
-    ACTION_TYPE_DELETE = 'delete'
     ACTION_TYPE_OTHER = 'other'
     ACTION_TYPE_SKIP = 'skip'
+    ACTION_TYPE_DEFAULT = ACTION_TYPE_CREATE_OR_UPDATE
+
+    # All defined action types used for importing, logging, and future functionality
+    ACTION_TYPES = [
+        ACTION_TYPE_DELETE,
+        ACTION_TYPE_CREATE_OR_UPDATE,
+        ACTION_TYPE_CREATE,
+        ACTION_TYPE_UPDATE,
+        ACTION_TYPE_RETIRE,
+        ACTION_TYPE_OTHER,
+        ACTION_TYPE_SKIP
+    ]
+
+    # Currently supported import action types
+    IMPORT_ACTION_TYPES = [
+        ACTION_TYPE_DELETE,
+        ACTION_TYPE_CREATE_OR_UPDATE,
+        ACTION_TYPE_CREATE,
+        ACTION_TYPE_UPDATE,
+    ]
+
+    # Constants for HTTP methods
+    HTTP_METHOD_POST = "POST"
+    HTTP_METHOD_PUT = "PUT"
+    HTTP_METHOD_DELETE = "DELETE"
+    HTTP_METHOD_GET = "GET"
+    HTTP_METHOD_HEAD = "HEAD"
+
+    # Constants for OCL resource definitions
+    OBJ_DEF_ATTR_HTTP_IMPORT_METHOD = "http_import_method"
+
+    # Constants for inline resource attributes
+    ATTR_RESOURCE_TYPE = 'type'
+    ATTR_ACTION = '__action'
 
     # Resource type definitions
+    # TODO: Transition to a single definition per resource that lives in OclDev
     obj_def = {
         oclconstants.OclConstants.RESOURCE_TYPE_ORGANIZATION: {
             "id_field": "id",
@@ -470,8 +509,11 @@ class OclFlexImporter(object):
                 "id", "company", "extras", "location", "name",
                 "public_access", "extras", "website"
             ],
-            "create_method": "POST",
-            "update_method": "PUT",
+            OBJ_DEF_ATTR_HTTP_IMPORT_METHOD: {
+                ACTION_TYPE_CREATE: HTTP_METHOD_POST,
+                ACTION_TYPE_UPDATE: HTTP_METHOD_PUT,
+                ACTION_TYPE_DELETE: HTTP_METHOD_DELETE
+            }
         },
         oclconstants.OclConstants.RESOURCE_TYPE_SOURCE: {
             "id_field": "id",
@@ -484,8 +526,11 @@ class OclFlexImporter(object):
                 "source_type", "custom_validation_schema", "public_access",
                 "default_locale", "supported_locales", "website", "extras", "external_id"
             ],
-            "create_method": "POST",
-            "update_method": "PUT",
+            OBJ_DEF_ATTR_HTTP_IMPORT_METHOD: {
+                ACTION_TYPE_CREATE: HTTP_METHOD_POST,
+                ACTION_TYPE_UPDATE: HTTP_METHOD_PUT,
+                ACTION_TYPE_DELETE: HTTP_METHOD_DELETE
+            }
         },
         oclconstants.OclConstants.RESOURCE_TYPE_COLLECTION: {
             "id_field": "id",
@@ -498,8 +543,11 @@ class OclFlexImporter(object):
                 "custom_validation_schema", "public_access", "default_locale", "supported_locales",
                 "website", "extras", "external_id"
             ],
-            "create_method": "POST",
-            "update_method": "PUT",
+            OBJ_DEF_ATTR_HTTP_IMPORT_METHOD: {
+                ACTION_TYPE_CREATE: HTTP_METHOD_POST,
+                ACTION_TYPE_UPDATE: HTTP_METHOD_PUT,
+                ACTION_TYPE_DELETE: HTTP_METHOD_DELETE
+            }
         },
         oclconstants.OclConstants.RESOURCE_TYPE_CONCEPT: {
             "id_field": "id",
@@ -511,8 +559,11 @@ class OclFlexImporter(object):
                 "id", "external_id", "concept_class", "datatype", "names",
                 "descriptions", "retired", "extras"
             ],
-            "create_method": "POST",
-            "update_method": "PUT",
+            OBJ_DEF_ATTR_HTTP_IMPORT_METHOD: {
+                ACTION_TYPE_CREATE: HTTP_METHOD_POST,
+                ACTION_TYPE_UPDATE: HTTP_METHOD_PUT,
+                ACTION_TYPE_DELETE: HTTP_METHOD_DELETE
+            }
         },
         oclconstants.OclConstants.RESOURCE_TYPE_MAPPING: {
             "id_field": "id",
@@ -524,8 +575,11 @@ class OclFlexImporter(object):
                 "id", "map_type", "from_concept_url", "to_source_url", "to_concept_url",
                 "to_concept_code", "to_concept_name", "extras", "external_id"
             ],
-            "create_method": "POST",
-            "update_method": "POST",
+            OBJ_DEF_ATTR_HTTP_IMPORT_METHOD: {
+                ACTION_TYPE_CREATE: HTTP_METHOD_POST,
+                ACTION_TYPE_UPDATE: HTTP_METHOD_POST,
+                ACTION_TYPE_DELETE: HTTP_METHOD_DELETE
+            }
         },
         oclconstants.OclConstants.RESOURCE_TYPE_REFERENCE: {
             "url_name": "references",
@@ -533,8 +587,11 @@ class OclFlexImporter(object):
             "has_source": False,
             "has_collection": True,
             "allowed_fields": ["data"],
-            "create_method": "PUT",
-            "update_method": None,
+            OBJ_DEF_ATTR_HTTP_IMPORT_METHOD: {
+                ACTION_TYPE_CREATE: HTTP_METHOD_POST,
+                ACTION_TYPE_UPDATE: None,
+                ACTION_TYPE_DELETE: HTTP_METHOD_DELETE
+            }
         },
         oclconstants.OclConstants.RESOURCE_TYPE_SOURCE_VERSION: {
             "id_field": "id",
@@ -544,8 +601,11 @@ class OclFlexImporter(object):
             "has_collection": False,
             "omit_resource_name_on_get": True,
             "allowed_fields": ["id", "external_id", "description", "released"],
-            "create_method": "POST",
-            "update_method": "POST",
+            OBJ_DEF_ATTR_HTTP_IMPORT_METHOD: {
+                ACTION_TYPE_CREATE: HTTP_METHOD_POST,
+                ACTION_TYPE_UPDATE: HTTP_METHOD_POST,
+                ACTION_TYPE_DELETE: HTTP_METHOD_DELETE
+            }
         },
         oclconstants.OclConstants.RESOURCE_TYPE_COLLECTION_VERSION: {
             "id_field": "id",
@@ -555,8 +615,11 @@ class OclFlexImporter(object):
             "has_collection": True,
             "omit_resource_name_on_get": True,
             "allowed_fields": ["id", "external_id", "description", "released"],
-            "create_method": "POST",
-            "update_method": "PUT",
+            OBJ_DEF_ATTR_HTTP_IMPORT_METHOD: {
+                ACTION_TYPE_CREATE: HTTP_METHOD_POST,
+                ACTION_TYPE_UPDATE: HTTP_METHOD_PUT,
+                ACTION_TYPE_DELETE: HTTP_METHOD_DELETE
+            }
         },
         oclconstants.OclConstants.RESOURCE_TYPE_USER: {
             "id_field": "username",
@@ -566,17 +629,18 @@ class OclFlexImporter(object):
             "has_collection": False,
             "allowed_fields": [
                 "username", "name", "email", "company", "location", "preferred_locale"],
-            "create_method": "POST",
-            "update_method": "POST",
+            OBJ_DEF_ATTR_HTTP_IMPORT_METHOD: {
+                ACTION_TYPE_CREATE: HTTP_METHOD_POST,
+                ACTION_TYPE_UPDATE: HTTP_METHOD_POST,
+                ACTION_TYPE_DELETE: HTTP_METHOD_DELETE
+            }
         }
     }
 
-
     def __init__(self, file_path='', input_list=None, api_url_root='', api_token='', limit=0,
                  test_mode=False, verbosity=1, do_update_if_exists=False, import_delay=0,
                  reference_batch_size=DEFAULT_REFERENCE_BATCH_SIZE):
         """ Initialize this object """
-
         self.input_list = input_list
         self.file_path = file_path
         if file_path:
@@ -599,7 +663,6 @@ class OclFlexImporter(object):
             'Content-Type': 'application/json'
         }
 
-
     def log(self, *args):
         """ Output log information """
         sys.stdout.write('[' + str(datetime.now()) + '] ')
@@ -609,7 +672,6 @@ class OclFlexImporter(object):
         sys.stdout.write('\n')
         sys.stdout.flush()
 
-
     def log_settings(self):
         """ Output log of the object settings """
         self.log("**** OCL IMPORT SCRIPT SETTINGS ****",
@@ -621,7 +683,6 @@ class OclFlexImporter(object):
                  ", Verbosity:", self.verbosity,
                  ", Import Delay: ", self.import_delay)
 
-
     def load_from_file(self, file_path):
         """ Load the OCL-formatted JSON file from the specified path """
         self.file_path = file_path
@@ -630,7 +691,6 @@ class OclFlexImporter(object):
             for json_line_raw in json_file:
                 self.input_list.append(json.loads(json_line_raw))
 
-
     def process(self):
         """
         Imports a JSON-lines file using OCL API
@@ -659,8 +719,8 @@ class OclFlexImporter(object):
             if self.limit > 0 and count >= self.limit:
                 break
             count += 1
-            if "type" in json_line_obj:
-                obj_type = json_line_obj.pop("type")
+            if OclFlexImporter.ATTR_RESOURCE_TYPE in json_line_obj:
+                obj_type = json_line_obj.pop(OclFlexImporter.ATTR_RESOURCE_TYPE)
                 if (obj_type == oclconstants.OclConstants.RESOURCE_TYPE_REFERENCE and
                         self.reference_batch_size):
                     self.process_reference_object(
@@ -670,14 +730,17 @@ class OclFlexImporter(object):
                     self.process_object(obj_type, json_line_obj)
                     num_processed += 1
                 else:
-                    message = "Unrecognized 'type' attribute '%s' for object: %s" % (
-                        obj_type, json_line_raw)
+                    # Invalid resource type attribute
+                    message = "Unrecognized '%s' attribute '%s' for object: %s" % (
+                        OclFlexImporter.ATTR_RESOURCE_TYPE, obj_type, json_line_raw)
                     self.import_results.add(action_type=self.ACTION_TYPE_SKIP, obj_type=obj_type,
                                             text=json_line_raw, message=message)
                     self.log('**** SKIPPING: %s' % message)
                     num_skipped += 1
             else:
-                message = "No 'type' attribute: %s" % json_line_raw
+                # Missing the resource type attribute
+                message = "No '%s' attribute: %s" % (
+                    OclFlexImporter.ATTR_RESOURCE_TYPE, json_line_raw)
                 self.import_results.add(
                     action_type=self.ACTION_TYPE_SKIP, text=json_line_raw, message=message)
                 self.log('**** SKIPPING: %s' % message)
@@ -694,13 +757,13 @@ class OclFlexImporter(object):
         return count
 
     def does_object_exist(self, obj_url, use_cache=True):
-        """ Returns whether a resource at the specified URL already exists """
+        """ Returns whether a resource at the specified URL already exists. """
 
-        # If resource existence cached, then just return True
+        # If we already know the resource exists, then just return True
         if use_cache and obj_url in self.cache_obj_exists and self.cache_obj_exists[obj_url]:
             return True
 
-        # Object existence not cached, so use API to check if it exists
+        # Object existence not cached, so use OCL API to check if it exists
         request_existence = requests.head(self.api_url_root + obj_url, headers=self.api_headers)
         if request_existence.status_code == requests.codes.ok:
             self.cache_obj_exists[obj_url] = True
@@ -714,8 +777,8 @@ class OclFlexImporter(object):
 
     def does_mapping_exist(self, obj_url, obj):
         """
-        Returns whether the specified mapping already exists --
-        Equivalent mapping must have matching source, from_concept, to_concept, and map_type
+        Returns whether a mapping already exists by finding a matching source from_concept,
+        to_concept, and map_type, or by finding a matching mapping ID (if provided).
         """
 
         # # Return false if correct fields not set
@@ -767,7 +830,6 @@ class OclFlexImporter(object):
 
         return False
 
-
     def does_reference_exist(self, obj_url, obj):
         """ Returns whether the specified reference already exists """
 
@@ -791,7 +853,6 @@ class OclFlexImporter(object):
         #     raise UnexpectedStatusCodeError(
         #         "GET " + self.api_url_root + obj_url,
         #         "Unexpected status code returned: " + str(request_existence.status_code))
-
         return False
 
     @staticmethod
@@ -814,7 +875,7 @@ class OclFlexImporter(object):
         return new_resources
 
     def process_reference_object(self, obj, batch_size=DEFAULT_REFERENCE_BATCH_SIZE):
-        """ Split list of references into batches and process """
+        """ Process list of references, optionally splitting into batches """
         batched_resources = OclFlexImporter.batch_reference_object(obj)
         if len(batched_resources) > 1:
             message = ('INFO: New reference request with %s expressions automatically '
@@ -827,8 +888,7 @@ class OclFlexImporter(object):
             self.process_object(obj_type, new_obj)
 
     def process_object(self, obj_type, obj):
-        """ Processes an individual document in the import file """
-
+        """ Processes an individual object in the import file """
         # Grab the object ID
         obj_id = ''
         if 'id_field' in self.obj_def[obj_type] and self.obj_def[obj_type]['id_field'] in obj:
@@ -951,7 +1011,8 @@ class OclFlexImporter(object):
             self.log("** Allowed Fields: **", json.dumps(obj))
             self.log("** Removed Fields: **", json.dumps(obj_not_allowed))
 
-        # Check if owner exists - at this point obj_url, obj_owner_url, and obj_repo_url are set
+        # Check if owner exists, for objects that have owners
+        # NOTE: At this point obj_url, obj_owner_url, and obj_repo_url are set
         if has_owner and obj_owner_url:
             try:
                 if self.does_object_exist(obj_owner_url):
@@ -974,7 +1035,7 @@ class OclFlexImporter(object):
                 self.log("** SKIPPING: %s" % message)
                 return
 
-        # Check if repository exists
+        # Check if repository exists, for objects that are saved in a repository
         if (has_source or has_collection) and obj_repo_url:
             try:
                 if self.does_object_exist(obj_repo_url):
@@ -1007,7 +1068,7 @@ class OclFlexImporter(object):
             else:
                 obj_already_exists = self.does_object_exist(obj_url)
         except UnexpectedStatusCodeError as err:
-            message = "Unexpected error occurred: %s, %s" % (err.expression, err.message)
+            message = "WARNING: Unexpected error occurred: %s, %s" % (err.expression, err.message)
             self.import_results.add(action_type=self.ACTION_TYPE_SKIP, obj_type=obj_type,
                                     obj_url=obj_url, obj_repo_url=obj_repo_url,
                                     obj_owner_url=obj_owner_url,
@@ -1015,7 +1076,7 @@ class OclFlexImporter(object):
             self.log("** SKIPPING: %s" % message)
             return
         if obj_already_exists and not self.do_update_if_exists:
-            message = "Object already exists at: %s%s" % (self.api_url_root, obj_url)
+            message = "INFO: Object already exists at: %s%s" % (self.api_url_root, obj_url)
             self.import_results.add(action_type=self.ACTION_TYPE_SKIP, obj_type=obj_type,
                                     obj_url=obj_url, obj_repo_url=obj_repo_url,
                                     obj_owner_url=obj_owner_url,
@@ -1029,65 +1090,75 @@ class OclFlexImporter(object):
             self.log("** INFO: Object does not exist so we'll create it at: %s%s" % (
                 self.api_url_root, obj_url))
 
-        # Create/update the object
+        # Perform the action: Create/update/delete the object
         try:
-            self.update_or_create(
-                obj_type=obj_type,
-                obj_id=obj_id,
-                obj_owner_url=obj_owner_url,
-                obj_repo_url=obj_repo_url,
-                obj_url=obj_url,
-                new_obj_url=new_obj_url,
-                obj_already_exists=obj_already_exists,
-                obj=obj, obj_not_allowed=obj_not_allowed,
+            self.import_resource(
+                obj_type=obj_type, obj_id=obj_id, obj_owner_url=obj_owner_url,
+                obj_repo_url=obj_repo_url, obj_url=obj_url, new_obj_url=new_obj_url,
+                obj_already_exists=obj_already_exists, obj=obj, obj_not_allowed=obj_not_allowed,
                 query_params=query_params)
         except requests.exceptions.HTTPError as err:
             self.log("ERROR: ", err)
 
+    def import_resource(self, obj_type='', obj_id='', obj_owner_url='', obj_repo_url='',
+                        obj_url='', new_obj_url='', obj_already_exists=False, obj=None,
+                        obj_not_allowed=None, query_params=None):
+        """ Imports a resource as a create, update, or delete request to the OCL API """
 
-    def update_or_create(self, obj_type='', obj_id='', obj_owner_url='',
-                         obj_repo_url='', obj_url='', new_obj_url='',
-                         obj_already_exists=False,
-                         obj=None, obj_not_allowed=None,
-                         query_params=None):
-        """ Posts an object to the OCL API as either an update or create """
-
-        # Determine which URL to use based on whether or not object already exists
-        if obj_already_exists:
-            method = self.obj_def[obj_type]['update_method']
-            url = obj_url
-            action_type = self.ACTION_TYPE_UPDATE
+        # Determine the action type and URL
+        if (OclFlexImporter.ATTR_ACTION in obj and
+                obj[OclFlexImporter.ATTR_ACTION] in OclFlexImporter.IMPORT_ACTION_TYPES):
+            action_type = obj[OclFlexImporter.ATTR_ACTION]
         else:
-            method = self.obj_def[obj_type]['create_method']
-            url = new_obj_url
-            action_type = self.ACTION_TYPE_NEW
+            action_type = OclFlexImporter.ACTION_TYPE_DEFAULT
+        if action_type == OclFlexImporter.ACTION_TYPE_CREATE_OR_UPDATE:
+            if obj_already_exists:
+                action_type = OclFlexImporter.ACTION_TYPE_UPDATE
+            else:
+                action_type = OclFlexImporter.ACTION_TYPE_CREATE
+        url = obj_url if obj_already_exists else new_obj_url
+
+        # Determine which HTTP method to use
+        obj_def = self.obj_def[obj_type]
+        if (OclFlexImporter.OBJ_DEF_ATTR_HTTP_IMPORT_METHOD in obj_def and
+                action_type in obj_def[OclFlexImporter.OBJ_DEF_ATTR_HTTP_IMPORT_METHOD]):
+            method = obj_def[OclFlexImporter.OBJ_DEF_ATTR_HTTP_IMPORT_METHOD][action_type]
+            if not method:
+                err_msg = 'Action type "%s" not supported for resource type "%s"' % (
+                    action_type, obj_type)
+                raise UnsupportedActionType(obj, err_msg)
+        else:
+            err_msg = 'No HTTP method specified for "%s" resource type and action type "%s"' % (
+                obj_type, action_type)
+            raise InvalidObjectDefinition(obj, err_msg)
 
         # Add query parameters (if provided)
         if query_params:
             url += '?' + urllib.urlencode(query_params)
 
-        # Get out of here if in test mode
-        if self.test_mode:
+        # Import the object
+        if not self.test_mode:
+            self.log(method, " ", self.api_url_root + url + '  ', json.dumps(obj))
+            if method == OclFlexImporter.HTTP_METHOD_POST:
+                request_result = requests.post(
+                    self.api_url_root + url, headers=self.api_headers, data=json.dumps(obj))
+            elif method == OclFlexImporter.HTTP_METHOD_PUT:
+                request_result = requests.put(
+                    self.api_url_root + url, headers=self.api_headers, data=json.dumps(obj))
+            elif method == OclFlexImporter.HTTP_METHOD_DELETE:
+                # NOTE: Deletes do not include the object in the request body
+                request_result = requests.delete(self.api_url_root + url, headers=self.api_headers)
+            self.log("STATUS CODE:", request_result.status_code)
+            self.log(request_result.headers)
+            self.log(request_result.text)
+            self.import_results.add(
+                obj_url=obj_url, action_type=action_type, obj_type=obj_type,
+                obj_repo_url=obj_repo_url, http_method=method, obj_owner_url=obj_owner_url,
+                status_code=request_result.status_code, text=json.dumps(obj),
+                message=request_result.text)
+            request_result.raise_for_status()
+        else:
             self.log("[TEST MODE] ", method, self.api_url_root + url + '  ', json.dumps(obj))
-            return
-
-        # Create or update the object
-        self.log(method, " ", self.api_url_root + url + '  ', json.dumps(obj))
-        if method == 'POST':
-            request_result = requests.post(self.api_url_root + url, headers=self.api_headers,
-                                           data=json.dumps(obj))
-        elif method == 'PUT':
-            request_result = requests.put(self.api_url_root + url, headers=self.api_headers,
-                                          data=json.dumps(obj))
-        self.log("STATUS CODE:", request_result.status_code)
-        self.log(request_result.headers)
-        self.log(request_result.text)
-        self.import_results.add(
-            obj_url=obj_url, action_type=action_type, obj_type=obj_type, obj_repo_url=obj_repo_url,
-            http_method=method, obj_owner_url=obj_owner_url, status_code=request_result.status_code,
-            text=json.dumps(obj), message=request_result.text)
-        request_result.raise_for_status()
-
 
     def find_nth(self, haystack, needle, n):
         """ Find nth occurrence of a substring within a string """
